from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Generic

from ..transform.types import Transformer
from ..types import ID, IdType, NameType, SourceType
from ._format import Format
from ._magic_dict import MagicDict
from .types import PlaceholdersTuple, PlaceholderTranslations, TranslatedIds

if TYPE_CHECKING:
    import pandas


class FormatApplier(Generic[NameType, SourceType, IdType]):
    """Application of :class:`.Format` specifications.

    This class converts raw translation data into ready-to-use dicts on the form ``{id: translation}``, where the
    translation is always a plain string.

    Args:
        translations: A :class:`~.PlaceholderTranslations` object returned by fetchers.
        transformer: Initialized :class:`.Transformer` instance.

    Raises:
        ValueError: If `default` is given and any placeholder names are missing.

    Examples:
        Basic usage.

        >>> data = {1999: "Sofia", 1991: "Richard", 1904: "Fred"}
        >>> translations = PlaceholderTranslations.from_dict("my-source", data)
        >>> applier = FormatApplier(translations)

        We used the simplified ``{id: name}`` to create the translation object above. Let's create the formats to use:

        >>> fmt = Format.parse("{id}:{name}")
        >>> default_fmt = Format.parse("<Failed: id={id!r}>")

        Using ``FormatApplier.__call__`` delegates to :meth:`apply`.

        >>> applier(fmt, default_fmt=default_fmt)
        {1999: '1999:Sofia', 1991: '1991:Richard', 1904: '1904:Fred'}

        The output may look like a regular ``dict``, but is actually a :class:`.MagicDict`.

        >>> magic_dict = applier(fmt, default_fmt=default_fmt)
        >>> type(magic_dict)
        <class 'id_translation.offline._magic_dict.MagicDict'>

        .. warning::
            The :class:`.MagicDict` is does **not** behave like a regular dict.

        You can, for instance, use ``__getitem__`` on unknown keys:

        >>> magic_dict = applier(fmt, default_fmt=default_fmt)
        >>> magic_dict[-1]
        '<Failed: id=-1>'

        See the :class:`.MagicDict` class documentation for more information.
    """

    def __init__(
        self,
        translations: PlaceholderTranslations[SourceType],
        *,
        transformer: Transformer[IdType] | None = None,
    ) -> None:
        self._translations = translations
        self._transformer = transformer

    def apply(
        self,
        fmt: Format,
        *,
        default_fmt: Format,
        placeholders: PlaceholdersTuple | None = None,
        default_fmt_placeholders: dict[str, Any] | None = None,
        enable_uuid_heuristics: bool = True,
    ) -> MagicDict[IdType]:
        """Translate IDs.

        .. note::

           This method does not accept strings. Use :meth:`.Format.parse` to convert raw formats.

        Args:
            fmt: Translation :class:`.Format` to use.
            placeholders: Tuple of placeholder names to include in the formatted output. If ``None``, use the
                intersection of :attr:`.placeholders` and :attr:`fmt.placeholders <.Format.placeholders>`.
            default_fmt: Alternative format for default translation.
            default_fmt_placeholders: Default placeholders, e.g. ``{'name': 'default name'}``.
            enable_uuid_heuristics: Improves matching when :py:class:`~uuid.UUID`-like IDs are in use.

        Returns:
            A dict ``{id: translation}``.

        Notes:
            This method is an alias of ``__call__``.
        """
        assert isinstance(fmt, Format), f"invalid {fmt=}"  # noqa: S101
        assert isinstance(default_fmt, Format), f"invalid {default_fmt=}"  # noqa: S101

        if placeholders is None:
            # Use as many placeholders as possible.
            predicate = {*self._placeholder_names, *self._translations.placeholder_aliases}.__contains__
            placeholders = tuple(filter(predicate, fmt.placeholders))

        fstring = fmt.fstring(placeholders, positional=True)
        real_translations = self._apply(fstring, placeholders)

        if default_fmt_placeholders is None:
            default_fmt_placeholders = {}

        partial = default_fmt.partial(default_fmt_placeholders)
        try:
            default_fstring = partial.fstring(positional=True)
        except KeyError as e:
            raise ValueError(
                f"All required placeholders except {{{ID}}} must be provided for {default_fmt=}:"
                f" {default_fmt.required_placeholders}."
            ) from e

        return MagicDict(real_translations, default_fstring, enable_uuid_heuristics, self._transformer)

    __call__ = apply

    def to_dict(self) -> dict[str, Sequence[Any]]:
        """Get the underlying data used for translations as a dict.

        Returns:
            A dict ``{placeholder: [values...]}``.
        """
        return self._translations.to_dict()

    def to_pandas(self) -> "pandas.DataFrame":
        """Get the underlying data used for translations as a :class:`pandas.DataFrame`."""
        from pandas import DataFrame  # noqa: PLC0415

        return DataFrame(self.to_dict()).convert_dtypes()

    @property
    def records(self) -> Sequence[Sequence[Any]]:
        """Records used by this instance; see :attr:`.PlaceholderTranslations.records`."""
        return self._translations.records

    def _apply(self, fstring: str, placeholders: PlaceholdersTuple) -> TranslatedIds[IdType]:
        """Apply fstring to all IDs.

        Args:
            fstring: A format string.
            placeholders: Keys needed for the fstring, in the order in which they appear.

        Returns:
            A dict ``{id: translation}``.
        """
        id_pos, records = self._translations.id_pos, self._translations.records

        if self._placeholder_names == placeholders:
            return {record[id_pos]: fstring.format(*record) for record in records}
        else:
            placeholders = tuple(self._translations.placeholder_aliases.get(p, p) for p in placeholders)
            pos = tuple(map(self._placeholder_names.index, placeholders))
            return {record[id_pos]: fstring.format(*(record[i] for i in pos)) for record in records}

    @property
    def _placeholder_names(self) -> PlaceholdersTuple:
        return self._translations.placeholders

    @property
    def source(self) -> SourceType:
        """Return translation source."""
        return self._translations.source

    @property
    def placeholders(self) -> list[str]:
        """List of placeholder names; see :attr:`.PlaceholderTranslations.placeholders`."""
        return list(self._translations.placeholders)

    @property
    def transformer(self) -> Transformer[IdType] | None:
        """Get the :class:`.Transformer` instance (or ``None``) used by this ``FormatApplier``."""
        return self._transformer

    def __len__(self) -> int:
        return len(self._translations.records)

    def __repr__(self) -> str:
        placeholders = self._placeholder_names
        source = self.source
        return f"{self.__class__.__name__}({len(self)} IDs, {placeholders=}, {source=})"



import logging
import warnings
from time import perf_counter
from typing import TYPE_CHECKING

from rics.strings import format_seconds as fmt_sec

from .. import logging as _logging
from ..mapping.exceptions import MappingError, MappingWarning
from ..mapping.matrix import ScoreMatrix
from ..mapping.types import UserOverrideFunction
from ..types import IdType, Names, NameToSource, NameType, NameTypes, SourceType, Translatable
from ._names import NamesTask

if TYPE_CHECKING:
    from .._translator import Translator
    from ..mapping import DirectionalMapping


LOGGER = logging.getLogger("id_translation.Translator.map")


class MappingTask(NamesTask[NameType, SourceType, IdType]):
    """Internal type; not part of the public API."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        task_id: int | None = None,
    ) -> None:
        super().__init__(
            caller,
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            task_id=task_id,
        )

        # Task outputs
        self._name_to_source: NameToSource[NameType, SourceType] | None = None

    @property
    def name_to_source(self) -> NameToSource[NameType, SourceType]:
        """Name-to-source mapping."""
        if self._name_to_source is None:
            self._name_to_source = self._map()
        return self._name_to_source

    @property
    def names_to_translate(self) -> list[NameType]:
        """Keys of :attr:`name_to_source`."""
        return list(self.name_to_source)

    def _map(self) -> NameToSource[NameType, SourceType]:
        start = perf_counter()

        values = self.mapper_input_names
        type_name = self.type_name
        names_from_user = self.names_from_user

        if names_from_user is not None and not self.mapper_input_names:
            msg = f"Translation aborted; no names to translate in {type_name}{self._format_params()}."
            warnings.warn(msg, MappingWarning, stacklevel=2)
            LOGGER.warning(msg, extra={"task_id": self.task_id})
            return {}

        sources = self.caller.sources
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                f"Begin name-to-source mapping of names={values} in {type_name} against {sources=}.",
                extra=dict(
                    task_id=self.task_id,
                    event_key=_logging.get_event_key(self.caller.map, "enter"),
                    translatable_type=self.full_type_name,
                    values=values,
                    candidates=sources,
                ),
            )

        result: DirectionalMapping[NameType, SourceType]
        result = self.caller.mapper.apply(values, sources, None, self.override_function, task_id=self.task_id)
        name_to_source = result.flatten()
        seconds = perf_counter() - start

        if LOGGER.isEnabledFor(logging.INFO):
            with_nones = {name: name_to_source.get(name) for name in self.mapper_input_names}
            display = with_nones if LOGGER.isEnabledFor(logging.DEBUG) else name_to_source
            LOGGER.info(
                f"Finished mapping of {len(name_to_source)}/{len(self.mapper_input_names)} names in {type_name} "
                f"in {fmt_sec(seconds)}: {display}.",
                extra=dict(
                    task_id=self.task_id,
                    event_key=_logging.get_event_key(self.caller.map, "exit"),
                    seconds=seconds,
                    translatable_type=self.full_type_name,
                    values=values,
                    candidates=sources,
                    mapping=with_nones,
                ),
            )

        if not name_to_source:
            tail = f"could not be mapped to {sources=}{self._format_params()}"

            if names_from_user is None:
                # We're only concerned with mapping here. The NamesTask manages the no-names-at-all case.
                if self.mapper_input_names:
                    msg = (
                        f"Translation aborted; none of the derived names {self.mapper_input_names}"
                        f" in the {type_name}-type data could be mapped to available {sources=}"
                        f"{self._format_params()}."
                    )
                    warnings.warn(msg, MappingWarning, stacklevel=2)
                    LOGGER.warning(msg, extra={"task_id": self.task_id})
                return {}

            unmapped = set(names_from_user).difference(name_to_source)
            if unmapped:
                # Fail if any of the explicitly given names fail to map to a source.
                msg = f"Required names {unmapped} {tail}."
                LOGGER.error(msg, extra={"task_id": self.task_id})
                raise MappingError(msg)

        if result.cardinality.many_right:  # pragma: no cover
            for value, candidates in result.left_to_right.items():
                if len(candidates) > 1:
                    raise MappingError(
                        f"Name-to-source mapping {name_to_source} is ambiguous; {value} -> {candidates}."
                        f"\nHint: Choose a different cardinality such that Mapper.cardinality.many_right is False."
                    )

        self.add_timing("map", seconds)
        return name_to_source

    def compute_scores(self) -> ScoreMatrix[NameType, SourceType]:
        """Compute name-to-source match scores."""
        return self.caller.mapper.compute_scores(
            self.mapper_input_names,
            self.caller.sources,
            override_function=self.override_function,
            task_id=self.task_id,
        )

    def _format_params(self) -> str:
        ignore_names = self.ignore_names
        override_function = self.override_function

        params = []
        if ignore_names is not None:
            params.append(f"{ignore_names=}")
        if override_function is not None:
            params.append(f"{override_function=}")
        return f". Parameters: ({', '.join(params)})" if params else ""



import logging
import pickle
import warnings
from collections.abc import Iterable
from copy import deepcopy
from datetime import timedelta
from time import perf_counter
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NoReturn,
    Self,
    Unpack,
    overload,
)

from rics.collections.dicts import InheritedKeysDict, MakeType
from rics.collections.misc import as_list
from rics.env.read import read_bool
from rics.misc import get_public_module, tname
from rics.paths import AnyPath, any_path_to_path
from rics.strings import format_seconds as fmt_sec

from . import logging as _logging
from ._tasks import MappingTask, TranslationTask
from .exceptions import ConfigurationChangedError, ConnectionStatusError, TranslationDisabledWarning
from .fetching import Fetcher
from .fetching.types import IdsToFetch
from .mapping import Mapper
from .mapping.exceptions import MappingError
from .mapping.matrix import ScoreMatrix
from .mapping.types import UserOverrideFunction
from .offline import Format, TranslationMap
from .offline.types import (
    FormatType,
    PlaceholderTranslations,
    SourcePlaceholderTranslations,
)
from .testing import TestFetcher, TestMapper
from .toml import TranslatorFactory, meta
from .transform.types import Transformers
from .translator_typing import CopyParams, FetcherTypes
from .types import (
    ID,
    CopyTranslatable,
    DictToId,
    DictToList,
    DictToOneTuple,
    DictToSet,
    DictToThreeTuple,
    DictToTwoTuple,
    DictToVarTuple,
    HasSources,
    IdType,
    IdTypes,
    InplaceTranslatable,
    Names,
    NameToSource,
    NameType,
    NameTypes,
    SourceType,
    Translatable,
)

LOGGER = logging.getLogger(__package__).getChild("Translator")


ID_TRANSLATION_DISABLED = "ID_TRANSLATION_DISABLED"

if TYPE_CHECKING:
    from uuid import UUID

    import pandas

    ID_TRANSLATION_PANDAS_IS_TYPED: bool = False


class Translator(Generic[NameType, SourceType, IdType], HasSources[SourceType]):
    r"""End-user interface for all translation tasks.

    See :meth:`.Translator.translate` for runtime configuration options. Any argument chosen when the ``Translator`` is
    created can be overridden with :meth:`.Translator.copy`. Use :meth:`.go_offline` to store translations in memory.

    Args:
        fetcher: A :class:`.Fetcher` or ready-to-use translations.
        fmt: String :class:`.Format` specification for translations.
        mapper: A :class:`~.mapping.Mapper` instance for binding names to sources.
        default_fmt: Alternative :class:`.Format` to use fallback translation of unknown IDs.
        default_fmt_placeholders: Shared and/or source-specific default placeholder values for unknown IDs. See
            :meth:`InheritedKeysDict.make() <rics.collections.dicts.InheritedKeysDict.make>` for details.
        enable_uuid_heuristics: Improves matching when :py:class:`~uuid.UUID`-like IDs are in use.
        transformers: A dict ``{source: transformer}`` of initialized :class:`.Transformer` instances.

    .. _translator-docstring-example:
    Examples:

        Basic usage. For a more complete use case, see the :ref:`dvdrental` example.

        Assume that we have data for people and animals as in the tables below::

            people:            animals:
                 id | name        id | name   | is_nice
              ------+---------   ----+--------+---------
               1991 | Richard      0 | Tarzan | false
               1999 | Sofia        1 | Morris | true
               1904 | Fred         2 | Simba  | true

        In most real cases we'd fetch this table from somewhere. In this case, however, there's so little data that we
        can simply enumerate the components needed for translation ourselves.

        >>> from id_translation import Translator
        >>> translation_data = {
        ...     "animals": {
        ...         "id": [0, 1, 2],
        ...         "name": ["Tarzan", "Morris", "Simba"],
        ...         "is_nice": [False, True, True],
        ...     },
        ...     "people": {1999: "Sofia", 1991: "Richard", 1904: "Fred"},
        ... }
        >>> fmt = "{id}:{name}[, nice={is_nice}]"
        >>> translator = Translator(translation_data, fmt=fmt)

        Since `people` only has columns `id` and `name`, we can use the simplified ``{id: name}`` data format. We're
        using the full format for `animals` since we have an additional `is_nice` column in this table.
        We didn't define a :class:`.Mapper`, so the column names must match exactly.

        >>> import pandas as pd
        >>> df = pd.DataFrame({"animals": [0, 2], "people": [1991, 1999]})
        >>> translator.translate(df)  # Returns a copy
                        animals        people
        0  0:Tarzan, nice=False  1991:Richard
        1    2:Simba, nice=True    1999:Sofia

        Check out the :ref:`translation-primer` to learn how this is done "under the hood".
    """

    def __init__(
        self,
        fetcher: FetcherTypes[NameType, SourceType, IdType] | None = None,
        fmt: FormatType = Format.DEFAULT,
        mapper: Mapper[NameType, SourceType, None] | None = None,
        default_fmt: FormatType = Format.DEFAULT_FAILED,
        default_fmt_placeholders: MakeType[SourceType, str, Any] | None = None,
        enable_uuid_heuristics: bool = False,
        transformers: Transformers[SourceType, IdType] | None = None,
    ) -> None:
        self._transformers = {} if transformers is None else transformers

        self._fmt = fmt if isinstance(fmt, Format) else Format(fmt)
        self._default_fmt_placeholders: InheritedKeysDict[SourceType, str, Any] | None
        self._default_fmt_placeholders, self._default_fmt = _handle_default(default_fmt, default_fmt_placeholders)
        self._enable_uuid_heuristics = enable_uuid_heuristics

        self._cached_tmap: TranslationMap[NameType, SourceType, IdType] = self._to_translation_map({})
        self._fetcher: Fetcher[SourceType, IdType]
        if fetcher is None:
            self._fetcher = TestFetcher([])  # No explicit sources
            if mapper:  # pragma: no cover
                warnings.warn(
                    f"Mapper instance {mapper} given; consider creating a TestFetcher([sources..])-instance manually.",
                    UserWarning,
                    stacklevel=2,
                )
            else:
                mapper = TestMapper()
            warnings.warn(
                "No fetcher given. Translation data will be automatically generated.",
                UserWarning,
                stacklevel=2,
            )
        elif isinstance(fetcher, Fetcher):
            self._fetcher = fetcher
        elif isinstance(fetcher, dict):
            self._cached_tmap = self._to_translation_map(
                {source: PlaceholderTranslations.make(source, pht) for source, pht in fetcher.items()}
            )
        elif isinstance(fetcher, TranslationMap):
            tmap = fetcher.copy()
            tmap.fmt = self._fmt
            tmap.default_fmt = self._default_fmt
            tmap.default_fmt_placeholders = self._default_fmt_placeholders
            self._cached_tmap = tmap
        else:
            raise TypeError(type(fetcher))  # pragma: no cover

        self._mapper: Mapper[NameType, SourceType, None] = mapper or Mapper()
        self._mapper.logger = logging.getLogger("id_translation.Translator.map")

        self._config_metadata: meta.ConfigMetadata | None = None
        self._translated_names: NameToSource[NameType, SourceType] | None = None

    @classmethod
    def from_config(
        cls,
        path: AnyPath,
        extra_fetchers: Iterable[AnyPath] = (),
    ) -> Self:
        """Create a :class:`.Translator` from TOML inputs.

        See :ref:`translator-config` for help.

        Args:
            path: Path to the main TOML configuration file.
            extra_fetchers: Paths to fetching configuration TOML files. If multiple fetchers are defined, they are
                ranked by input order. If a fetcher defined in the main configuration, it will be prioritized (rank=0).

        Returns:
            A new :class:`.Translator` instance with a :attr:`config_metadata` attribute.
        """
        # docs: https://id-translation.readthedocs.io/en/stable/documentation/translator-config.html
        factory: TranslatorFactory[NameType, SourceType, IdType] = TranslatorFactory(path, extra_fetchers, cls)
        return factory.create()  # type: ignore[return-value]

    @property
    def config_metadata(self) -> meta.ConfigMetadata:
        """Return :func:`~Translator.from_config` initialization :class:`metadata <.ConfigMetadata>`."""
        if self._config_metadata is None:
            raise ValueError("Not created using Translator.from_config()")  # pragma: no cover
        return self._config_metadata

    def initialize_sources(self, task_id: int | None = None, *, force: bool = False) -> Self:
        """Perform source discovery (fetcher initialization).

        This method does nothing if the ``Translator`` isn't :attr:`online`.

        Args:
            task_id: Used for logging.
            force: If ``True``, perform full discovery even if sources are already known.

        Returns:
            Self, for chained assignment.
        """
        if self.online:
            if task_id is None:
                task_id = _logging.generate_task_id()
            self.fetcher.initialize_sources(task_id, force=force)
        return self

    def copy(self, **overrides: Unpack[CopyParams[NameType, SourceType, IdType]]) -> Self:
        """Make a copy of this :class:`.Translator`.

        Args:
            overrides: Keyword arguments to use when instantiating the copy. Options that aren't given will be taken
                from the current instance. See the :class:`Translator` class documentation for possible choices.

        Returns:
            A copy of this :class:`.Translator` with `overrides` applied.

        Notes:
            User types are copied using :func:`copy.deepcopy`.
        """
        kwargs: dict[str, Any] = {
            "fmt": self.fmt,
            "default_fmt": self.default_fmt,
            "enable_uuid_heuristics": self.enable_uuid_heuristics,
            **overrides,
        }

        if "mapper" not in kwargs:  # pragma: no cover
            kwargs["mapper"] = self.mapper.copy()
        if "default_fmt_placeholders" not in kwargs:
            kwargs["default_fmt_placeholders"] = self._default_fmt_placeholders
        if "fetcher" not in kwargs:
            fetcher: Fetcher[SourceType, IdType] | TranslationMap[NameType, SourceType, IdType]
            if self.online:
                fetcher = self.fetcher
                try:
                    fetcher = deepcopy(fetcher)
                except TypeError as e:
                    msg = (
                        f"Failed to clone fetcher (TypeError: {e}). Caller instance will be reused."
                        "\nHnt: To suppress this warning: Translator.copy(fetcher=Translator.fetcher)"
                    )
                    warnings.warn(msg, category=UserWarning, stacklevel=2)
            else:
                fetcher = self._cached_tmap.copy()
            kwargs["fetcher"] = fetcher
        if "transformers" in kwargs:
            transformers = []
            for t in self.transformers:
                try:
                    tc = deepcopy(t)
                    transformers.append(tc)
                except TypeError as e:
                    msg = (
                        f"Failed to clone {t!r}. (TypeError: {e}). Caller instance will be reused."
                        "\nHint: To suppress this warning: Translator.copy(transformers=Translator.transformers)"
                    )
                    warnings.warn(msg, category=UserWarning, stacklevel=2)
                    transformers = [*self.transformers]
            kwargs["transformers"] = transformers

        return type(self)(**kwargs)

    if TYPE_CHECKING:

        @overload
        def translate(
            self,
            translatable: InplaceTranslatable[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            # https://github.com/python/mypy/issues/7333#issuecomment-788255229
            copy: Literal[False],
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> None: ...

        @overload
        def translate(
            self,
            translatable: CopyTranslatable[IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            # https://github.com/python/mypy/issues/7333#issuecomment-788255229
            copy: Literal[False],
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> NoReturn: ...

        @overload
        def translate(
            self,
            translatable: IdTypes,
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> str: ...

        @overload
        def translate(
            self,
            translatable: list[IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> list[str]: ...

        # This doesn't seem to work; nested generic type issue?
        # TODO Need Higher-Kinded TypeVars?
        # @overload
        # def translate(
        #     self,
        #     translatable: List[List[IdType]],
        #     names: Union[NameTypes[NameType], NameToSource[NameType, SourceType]] = None,
        #     *,
        #     copy: Literal[True] = True,
        #     ignore_names: Names[NameType] | None = None,
        #     override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        #     max_fails: float = 1.0,
        #     # Translation specification
        #     reverse: bool = False,
        #     fmt: FormatType | None = None,
        # ) -> List[List[str]]:
        #     ...

        @overload
        def translate(
            self,
            translatable: set[IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> set[str]: ...

        @overload
        def translate(
            self,
            # This is not correct, but using the TypeVar (which the user may bind to get proper typing) doesn't work
            # at this time (python 3.11.3, mypy 1.5.1). Higher-Kinded TypeVars might solve this.
            # TODO: Higher-Kinded TypeVars
            translatable: DictToId[NameType, int] | DictToId[NameType, str] | DictToId[NameType, UUID],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToId[NameType, str]: ...

        @overload
        def translate(
            self,
            translatable: DictToSet[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToSet[NameType, str]: ...

        @overload
        def translate(
            self,
            translatable: DictToList[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToList[NameType, str]: ...

        @overload
        def translate(  # Overlaps with DictToVarTuple
            self,
            translatable: DictToOneTuple[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToOneTuple[NameType, str]: ...

        @overload
        def translate(  # Overlaps with DictToVarTuple
            self,
            translatable: DictToTwoTuple[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToTwoTuple[NameType, str]: ...

        @overload
        def translate(  # Overlaps with DictToVarTuple
            self,
            translatable: DictToThreeTuple[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToThreeTuple[NameType, str]: ...

        @overload
        def translate(
            self,
            translatable: DictToVarTuple[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> DictToVarTuple[NameType, str]: ...

        @overload
        def translate(
            self,
            translatable: tuple[IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> tuple[str]: ...

        @overload
        def translate(
            self,
            translatable: tuple[IdType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> tuple[str, str]: ...

        @overload
        def translate(
            self,
            translatable: tuple[IdType, IdType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> tuple[str, str, str]: ...

        @overload
        def translate(
            self,
            translatable: tuple[IdType, ...],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            copy: Literal[True] = True,
            ignore_names: Names[NameType] | None = None,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            # Translation specification
            reverse: bool = False,
            fmt: FormatType | None = None,
        ) -> tuple[str, ...]: ...

        if ID_TRANSLATION_PANDAS_IS_TYPED:
            # pandas-stubs or similar

            @overload
            def translate(
                self,
                translatable: "pandas.DataFrame",
                names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
                *,
                copy: Literal[True] = True,
                ignore_names: Names[NameType] | None = None,
                override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
                max_fails: float = 1.0,
                # Translation specification
                reverse: bool = False,
                fmt: FormatType | None = None,
            ) -> "pandas.DataFrame": ...

            @overload
            def translate(
                self,
                translatable: "pandas.Series[Any]",
                names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
                *,
                copy: Literal[True] = True,
                ignore_names: Names[NameType] | None = None,
                override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
                max_fails: float = 1.0,
                # Translation specification
                reverse: bool = False,
                fmt: FormatType | None = None,
            ) -> "pandas.Series[str]": ...

            @overload
            def translate(
                self,
                translatable: "pandas.Index[Any]",
                names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
                *,
                copy: Literal[True] = True,
                ignore_names: Names[NameType] | None = None,
                override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
                max_fails: float = 1.0,
                # Translation specification
                reverse: bool = False,
                fmt: FormatType | None = None,
            ) -> "pandas.Index[str]": ...

            @overload
            def translate(
                self,
                translatable: "pandas.DataFrame" | "pandas.Series[Any]",
                names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
                *,
                copy: Literal[False],
                ignore_names: Names[NameType] | None = None,
                override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
                max_fails: float = 1.0,
                # Translation specification
                reverse: bool = False,
                fmt: FormatType | None = None,
            ) -> None: ...

            @overload
            def translate(
                self,
                translatable: "pandas.Index[Any]",
                names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
                *,
                # https://github.com/python/mypy/issues/7333#issuecomment-788255229
                copy: Literal[False],
                ignore_names: Names[NameType] | None = None,
                override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
                max_fails: float = 1.0,
                # Translation specification
                reverse: bool = False,
                fmt: FormatType | None = None,
            ) -> NoReturn: ...

        @overload
        def translate(
            self,
            translatable: Translatable[NameType, IdType],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            ignore_names: Names[NameType] | None = None,
            copy: Literal[True] = True,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            reverse: Literal[True] = True,
            fmt: FormatType | None = None,
        ) -> Translatable[NameType, str]: ...

        @overload
        def translate(
            self,
            translatable: Translatable[NameType, str],
            names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
            *,
            ignore_names: Names[NameType] | None = None,
            copy: Literal[True] = True,
            override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
            max_fails: float = 1.0,
            reverse: Literal[True] = True,
            fmt: FormatType | None = None,
        ) -> Translatable[NameType, IdType]: ...

    def translate(
        self,
        translatable: Translatable[NameType, IdType],
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        copy: bool = True,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        max_fails: float = 1.0,
        reverse: bool = False,
        fmt: FormatType | None = None,
    ) -> Translatable[NameType, str] | None:
        """Translate IDs to human-readable strings.

        Simplified process:
            1. The :attr:`map` method performs name-to-source mapping (see :class:`~.DirectionalMapping`).
            2. The :attr:`fetch` method extracts IDs to translate and retrieves data (see :class:`.TranslationMap`).
            3. Finally, the :attr:`translate` method applies the translations and returns to the caller.

        See the :ref:`translation-primer` page for a detailed process description.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.

        Args:
            translatable: A data structure to translate.
            names: Explicit names to translate. Derive from `translatable` if ``None``. Alternatively, you may pass a
                ``dict`` on the form ``{name_in_translatable: source_to_use}``.
            ignore_names: Names **not** to translate, or a predicate ``(NameType) -> bool``.
            copy: If ``False``, translate in-place and return ``None``.
            override_function: A callable ``(name, sources, ids) -> Source | None``. See :meth:`.Mapper.apply` for details.
            max_fails: The maximum fraction of IDs for which translation may fail. 1=disabled.
            reverse: If ``True``, perform translations back to IDs. Offline mode only.
            fmt: A :class:`format string <.Format>` such as **'{id}:{name}'** use. Default is :attr:`.Translator.fmt`.

        Returns:
            A translated copy of `translatable` if ``copy=True``, otherwise ``None``.

        Examples:
            Manual `name-to-source <../documentation/translation-primer.html#name-to-source-mapping>`__ mapping with a
            temporary name-only :class:`.Format`.

            ..
               # Hidden setup code
               >>> translator = Translator({"animals": {"id": [2], "name": ["Simba"]}})

            >>> n2s = {"lions": "animals", "big_cats": "animals"}
            >>> translator.translate({"lions": 2, "big_cats": 2}, names=n2s, fmt="{name}")
            {'lions': 'Simba', 'big_cats': 'Simba'}

            Name mappings must be complete; any name not present in the keys will be ignored (left as-is).

        Raises:
            UntranslatableTypeError: If ``type(translatable)`` cannot be translated.
            MissingNamesError: If `names` are not given and cannot be derived from `translatable`.
            MappingError: If any required (explicitly given) names fail to map to a source.
            MappingError: If name-to-source mapping is ambiguous.
            ValueError: If `max_fails` is not a valid fraction.
            TooManyFailedTranslationsError: If translation fails for more than `max_fails` of IDs.
            ConnectionStatusError: If ``reverse=True`` while the :class:`.Translator` is online.
            UserMappingError: If `override_function` returns a source which is not known, and
                ``mapper.on_unknown_user_override != 'ignore'``.

        See Also:
            The :envvar:`ID_TRANSLATION_DISABLED` variable.
        """
        if read_bool(ID_TRANSLATION_DISABLED):
            message = f"Translation aborted; {ID_TRANSLATION_DISABLED} is set."
            LOGGER.warning(message)
            warnings.warn(message, category=TranslationDisabledWarning, stacklevel=2)
            return translatable if copy else None  # Return unchanged; breaks typing.

        if self.online and reverse:  # pragma: no cover
            raise ConnectionStatusError("Reverse translation cannot be performed online.")

        task_id = _logging.generate_task_id()
        self.initialize_sources(task_id)

        task: TranslationTask[NameType, SourceType, IdType] = TranslationTask(
            self,
            translatable,
            self._fmt if fmt is None else Format.parse(fmt),
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            copy=copy,
            max_fails=max_fails,
            reverse=reverse,
            enable_uuid_heuristics=self._enable_uuid_heuristics,
            event_key=f"{type(self).__name__}.{self.translate.__name__}",
            task_id=task_id,
        )
        task.log_key_event_enter()

        translation_map = self._get_updated_tmap(task)
        if not translation_map:
            # Return unchanged; this is technically against the API spec. If the user has required translation to
            # success through configuration, exceptions will be raised elsewhere. I don't know how to express this using
            # the Python type system.
            return translatable if copy else None

        task.verify(translation_map)

        ans: Translatable[NameType, str] | None = task.insert(translation_map)

        self._translated_names = dict(task.name_to_source)
        task.finished()
        task.log_key_event_exit()
        return ans

    @overload
    def translated_names(self, with_source: Literal[True]) -> NameToSource[NameType, SourceType]: ...

    @overload
    def translated_names(self, with_source: Literal[False] = False) -> list[NameType]: ...

    def translated_names(self, with_source: bool = True) -> NameToSource[NameType, SourceType] | list[NameType]:
        """Return the names that were translated by the most recent :meth:`.translate`-call.

        Args:
            with_source: If ``True``, return a dict ``{name: source}`` instead of a list.

        Returns:
            Recent names translated by this :class:`.Translator`, in **arbitrary** order.

        Raises:
            ValueError: If no names have been translated using this :class:`.Translator`.
        """
        if self._translated_names is None:
            raise ValueError("No names have been translated using this Translator.")
        return dict(self._translated_names) if with_source else list(self._translated_names)

    def map(
        self,
        translatable: Translatable[NameType, IdType],
        names: NameTypes[NameType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
    ) -> NameToSource[NameType, SourceType]:
        """Map names to translation sources.

        Args:
            translatable: A data structure to map names for.
            names: Explicit names to translate. Derive from `translatable` if ``None``.
            ignore_names: Names **not** to translate, or a predicate ``(NameType) -> bool``.
            override_function: A callable ``(name, sources, ids) -> Source | None``. See
                :meth:`Mapper.apply <.mapping.Mapper.apply>` for details.

        Returns:
            A mapping of names to translation sources. Returns ``None`` if mapping failed.

        Raises:
            MissingNamesError: If `names` are not given and cannot be derived from `translatable`.
            MappingError: If any required (explicitly given) names fail to map to a source.
            MappingError: If name-to-source mapping is ambiguous.
            UserMappingError: If `override_function` returns a source which is not known, and
                ``mapper.on_unknown_user_override != 'ignore'``.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """
        task_id = _logging.generate_task_id()
        self.initialize_sources(task_id)

        return MappingTask(
            self,
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            task_id=task_id,
        ).name_to_source

    def map_scores(
        self,
        translatable: Translatable[NameType, IdType],
        names: NameTypes[NameType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
    ) -> ScoreMatrix[NameType, SourceType]:
        """Returns raw match scores for name-to-source mapping. See :meth:`map` for details."""
        return MappingTask(
            self,
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
        ).compute_scores()

    @property
    def sources(self) -> list[SourceType]:
        """A list of known sources names.

        Sources are determines either by the :attr:`.fetcher` or the :attr:`.cache`.
        """
        return list(self.placeholders)

    @property
    def placeholders(self) -> dict[SourceType, list[str]]:
        """A dict ``source: [placeholders..]}``.

        Placeholders shown here are the names as they appear **in the source**.
        """
        return self._fetcher.placeholders if self.online else self._cached_tmap.placeholders

    @property
    def fmt(self) -> Format:
        """Main translation :class:`.Format` for this ``Translator`` instance."""
        return self._fmt

    @property
    def default_fmt(self) -> Format | None:
        """Alternative translation :class:`.Format`, used for unknown IDs."""
        return self._default_fmt

    @property
    def enable_uuid_heuristics(self) -> bool:
        """Improves matching when :py:class:`~uuid.UUID`-like IDs are in use."""
        return self._enable_uuid_heuristics

    @property
    def online(self) -> bool:
        """Return connectivity status. If ``False``, no new translations may be fetched."""
        return hasattr(self, "_fetcher")

    @property
    def fetcher(self) -> Fetcher[SourceType, IdType]:
        """Return the :class:`.Fetcher` instance used to retrieve translations."""
        if not self.online:
            raise ConnectionStatusError(
                "Cannot fetch new translations.\nHint: Use the Translator.cache-property to access the data."
            )

        return self._fetcher

    @property
    def mapper(self) -> Mapper[NameType, SourceType, None]:
        """Return the :class:`.Mapper` instance used for name-to-source binding."""
        return self._mapper

    @property
    def cache(self) -> TranslationMap[NameType, SourceType, IdType]:
        """Return a :class:`.TranslationMap` of cached translations."""
        return self._cached_tmap

    @classmethod
    def load_persistent_instance(
        cls,
        cache_dir: AnyPath,
        config_path: AnyPath,
        extra_fetchers: Iterable[AnyPath] = (),
        max_age: str | timedelta | None = "12h",
        on_config_changed: Literal["raise", "recreate"] = "recreate",
    ) -> Self:
        """Load or create a persistent :attr:`~.Fetcher.fetch_all`-instance.

        Instances are created, stored and loaded as determined by a metadata file located in the given `cache_dir`. A
        new :class:`.Translator` will be created if:

        * There is no `'metadata'` file, or
        * the original :class:`.Translator` is too old (see `max_age`), or
        * the current configuration -- as defined by ``(config_path, extra_fetchers, clazz)`` -- has changed in such a
          way that it is no longer equivalent configuration used to create the original :class:`.Translator`. For
          details, see :class:`~.toml.meta.ConfigMetadata`.

        .. warning:: This method is **not** thread safe.

        Args:
            cache_dir: Root directory where the cached translator and associated metadata is stored.
            config_path: Path to the main TOML configuration file.
            extra_fetchers: Paths to fetching configuration TOML files. If multiple fetchers are defined, they are
                ranked by input order. If a fetcher defined in the main configuration, it will be prioritized (rank=0).
            max_age: The maximum age of the cached :class:`.Translator` before it must be recreated. Pass zero to force
                recreation, or ``None`` to ignore.
            on_config_changed: One of ``raise|recreate``. If ``'raise'``, crash instead of creating a new instance
                if the configuration (as determined by `config_path` and `extra_fetchers`) has changed.

        Returns:
            A new or cached :class:`.Translator` instance with a :attr:`config_metadata` attribute.

        Raises:
            ConfigurationChangedError: If the configuration has changed and ``on_config_mismatch='raise'``.

        See Also:
             The :meth:`from_config` method, which will read the `config_path`.
        """
        path = any_path_to_path(config_path)
        cache_dir = any_path_to_path(cache_dir).expanduser().absolute()
        cache_dir.mkdir(parents=True, exist_ok=True)

        metadata_path = cache_dir / "metadata.json"
        cache_path = cache_dir / "translator.pkl"

        extra_fetcher_paths: list[str] = list(map(str, extra_fetchers))

        metadata = meta.ConfigMetadata.from_toml_paths(
            str(path),
            extra_fetcher_paths,
            clazz=cls,
        )
        use_cached, reason, reason_type = metadata.use_cached(metadata_path, max_age)
        if use_cached:
            LOGGER.info(f"Reuse existing Translator; {reason}. Cache dir: '{cache_dir}'.")
            return cls.restore(cache_path)

        if reason_type == "metadata-changed" and on_config_changed.lower() == "raise":
            raise ConfigurationChangedError(reason)

        LOGGER.info(f"Create new Translator; {reason}. Cache dir: '{cache_dir}'.")
        translator = cls.from_config(path, extra_fetcher_paths)
        translator.go_offline(path=cache_path)
        metadata_path.write_text(translator.config_metadata.to_json())
        return translator

    @classmethod
    def restore(cls, path: AnyPath) -> Self:
        """Restore a serialized :class:`.Translator`.

        Args:
            path: Path to a serialized :class:`.Translator`.

        Returns:
            A :class:`.Translator`.

        Raises:
            TypeError: If the object at `path` is not a :class:`.Translator` or a subtype thereof.

        See Also:
            The :meth:`go_offline` method.
        """
        full_path = any_path_to_path(path).expanduser()
        with full_path.open("rb") as f:
            ans = pickle.load(f)  # noqa: S301

        if type(ans) is not cls:  # pragma: no cover
            raise TypeError(f"Serialized object at at '{full_path}' is a {type(ans)}, not {cls}.")

        if LOGGER.isEnabledFor(logging.DEBUG):
            extra = "" if ans._config_metadata is None else f" with {ans.config_metadata}"
            LOGGER.debug(f"Deserialized {ans}{extra}.")

        return ans

    def go_offline(
        self,
        translatable: Translatable[NameType, IdType] | None = None,
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        max_fails: float = 1.0,
        fmt: FormatType | None = None,
        path: AnyPath | None = None,
    ) -> Self:
        """Retrieve and store translations in memory.

        .. warning::

           The translator will be disconnected. No new translations may be fetched after this method returns.

        Subsequent calls to this method will return immediately.

        Args:
            translatable: Data from which IDs to fetch will be extracted. Fetch all IDs if ``None``.
            names: Explicit names to translate. Derive from `translatable` if ``None``.
            ignore_names: Names **not** to translate, or a predicate ``(NameType) -> bool``.
            override_function: A callable ``(name, sources, ids) -> Source | None``. See :meth:`.Mapper.apply`
                for details.
            max_fails: The maximum fraction of IDs for which translation may fail. 1=disabled.
            fmt: A :class:`format string <.Format>` such as **'{id}:{name}'** use. Default is :attr:`.Translator.fmt`.
            path: If given, serialize the :class:`.Translator` to disk after retrieving data.

        Returns:
            Self, for chained assignment.

        Raises:
            ForbiddenOperationError: If :meth:`.Fetcher.fetch_all` is disabled and ``translatable=None``.
            MappingError: If :meth:`map` fails (only when `translatable` is given).

        Notes:
            The :class:`.Translator` is guaranteed to be :func:`~rics.misc.serializable` once offline. Fetchers often
            aren't as they require things like database connections to function.

        See Also:
            The :meth:`restore` method.
        """
        if not self.online:
            return self

        start = perf_counter()
        task_id = _logging.generate_task_id(start)

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                msg=f"Begin going offline with {len(self.sources)} sources provided by: {self.fetcher}",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.go_offline, "enter"),
                    # Task-specific
                    path=None if path is None else str(path),
                    translatable_type=None
                    if translatable is None
                    else get_public_module(type(translatable), resolve_reexport=True, include_name=True),
                ),
            )

        translation_map = self._user_fetch(
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            max_fails=max_fails,
            fmt=fmt,
            func=self.go_offline.__qualname__,
            task_id=task_id,
        )
        self.fetcher.close()
        del self._fetcher
        self._cached_tmap = translation_map

        if path:
            path = any_path_to_path(path).expanduser()
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("wb") as f:
                pickle.dump(self, f)

            cls = tname(self, include_module=True)
            LOGGER.info(
                f"Serialized '{cls}' of size {path.stat().st_size / 2**20:.2g} MiB at path='{path}'.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.go_offline, "serialize"),
                    # Task-specific
                    path=str(path),
                    translatable_type=None
                    if translatable is None
                    else get_public_module(type(translatable), resolve_reexport=True, include_name=True),
                ),
            )

        if LOGGER.isEnabledFor(logging.INFO):
            seconds = perf_counter() - start
            LOGGER.info(
                f"Went offline with {len(translation_map.sources)} sources in {fmt_sec(seconds)}: {translation_map}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.go_offline, "enter"),
                    seconds=seconds,
                    # Task-specific
                    path=None if path is None else str(path),
                    translatable_type=None
                    if translatable is None
                    else get_public_module(type(translatable), resolve_reexport=True, include_name=True),
                ),
            )

        return self

    def fetch(
        self,
        translatable: Translatable[NameType, IdType] | None = None,
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        max_fails: float = 1.0,
        fmt: FormatType | None = None,
    ) -> TranslationMap[NameType, SourceType, IdType]:
        """Fetch translations.

        Calling ``fetch`` without arguments will perform a :meth:`.Fetcher.fetch_all` -operation, without going offline.

        The returned :class:`.TranslationMap` may be converted to native types with :meth:`.TranslationMap.to_dicts`.

        Args:
            translatable: A data structure to translate.
            names: Explicit names to translate. Derive from `translatable` if ``None``. Alternatively, you may pass a
                ``dict`` on the form ``{name_in_translatable: source_to_use}``.
            ignore_names: Names **not** to translate, or a predicate ``(NameType) -> bool``.
            override_function: A callable ``(name, sources, ids) -> Source | None``. See :meth:`.Mapper.apply`
                for details.
            max_fails: The maximum fraction of IDs for which translation may fail. 1=disabled.
            fmt: A :class:`format string <.Format>` such as **'{id}:{name}'** use. Default is :attr:`.Translator.fmt`.

        Returns:
            A :class:`.TranslationMap`.

        Raises:
            ConnectionStatusError: If disconnected from the fetcher, i.e. not :attr:`online`.

        Examples:
            Using the returned :class:`.TranslationMap` class.

            ..
               # Hidden setup code
               >>> from id_translation.fetching import MemoryFetcher
               >>> translation_data = {
               ...     "animals": {"id": [0, 1, 2], "name": ["Tarzan", "Morris", "Simba"]},
               ...     "people": {"id": [1999, 1991], "name": ["Sofia", "Richard"]},
               ... }
               >>> translator = Translator(MemoryFetcher(translation_data))

            >>> translation_map = translator.fetch()
            >>> translation_map
            TranslationMap('animals': 3 IDs, 'people': 2 IDs)

            **Convert to finished translations.**

            * :meth:`.TranslationMap.to_translations` → ``{source: MagicDict}``, where a :class:`.MagicDict` is similar
              to a regular ``dict[IdType, str]``-type dict.

            >>> people = translation_map.to_translations()["people"]
            >>> people
            {1999: '1999:Sofia', 1991: '1991:Richard'}

            .. warning::

               The :class:`.MagicDict` class is used internally and has a few important differences from the built-in
               type. Please refer to the :class:`.MagicDict` class documentation for details.

            To convert to a :class:`.MagicDict` to a regular ``dict``, simply use the dict constructor:

            >>> dict(people)
            {1999: '1999:Sofia', 1991: '1991:Richard'}

            **Convert to raw translation data.**

            * :meth:`.TranslationMap.to_pandas` → ``{source: DataFrame}``
            * :meth:`.TranslationMap.to_dicts` → ``{source: {placeholder: [values...]}}``

            >>> translation_map.to_dicts()["people"]
            {'id': [1999, 1991], 'name': ['Sofia', 'Richard']}
        """
        task_id = _logging.generate_task_id()
        self.initialize_sources(task_id)

        return self._user_fetch(
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            max_fails=max_fails,
            fmt=fmt,
            func=self.fetch.__qualname__,
            task_id=task_id,
        )

    def _user_fetch(
        self,
        translatable: Translatable[NameType, IdType] | None = None,
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        max_fails: float = 1.0,
        fmt: FormatType | None = None,
        func: str,
        task_id: int | None = None,
    ) -> TranslationMap[NameType, SourceType, IdType]:
        fmt = self._fmt if fmt is None else Format.parse(fmt)

        if translatable is None:
            if all(p is None for p in (names, ignore_names, override_function)):
                source_translations = self._fetch(None, task_id=task_id)
            else:
                dummy = {source: None for source in self.sources}
                sources = self.map(dummy, names, ignore_names=ignore_names, override_function=override_function)
                source_translations = self._fetch(set(sources.values()), task_id=task_id)
            translation_map = self._to_translation_map(source_translations, fmt=fmt)

            if names is None:
                pass  # Callers must perform mapping unless name=source.
            elif isinstance(names, dict):
                translation_map.name_to_source = names
            else:
                translation_map.name_to_source = self.mapper.apply(
                    as_list(names),
                    translation_map.sources,
                    override_function=override_function,
                    task_id=task_id,
                ).flatten()
        else:
            task = TranslationTask(
                self,
                translatable,
                self._fmt,
                names,
                ignore_names=ignore_names,
                max_fails=max_fails,
                enable_uuid_heuristics=self._enable_uuid_heuristics,
                event_key=func,
                task_id=task_id,
            )
            if not task.name_to_source:
                msg = f"No names in the {tname(translatable, prefix_classname=True)!r}-type data were mapped."
                raise MappingError(msg)

            translation_map = self._get_updated_tmap(task, force_fetch=True)
            if LOGGER.isEnabledFor(logging.DEBUG):
                not_fetched = set(self.fetcher.sources).difference(translation_map.sources)
                LOGGER.debug(f"Available sources {not_fetched} were not fetched.")

            task.verify(translation_map)

        return translation_map

    def _get_updated_tmap(
        self,
        task: TranslationTask[NameType, SourceType, IdType],
        force_fetch: bool = False,
    ) -> TranslationMap[NameType, SourceType, IdType]:
        """Get an updated translation map."""
        name_to_source = task.name_to_source
        if not name_to_source:
            return self._to_translation_map({})  # Nothing to translate.

        translation_map = self._execute_fetch(task) if (force_fetch or not self.cache) else self.cache

        translation_map.enable_uuid_heuristics = task.enable_uuid_heuristics
        translation_map.fmt = task.fmt
        translation_map.name_to_source = task.name_to_source  # Update
        return translation_map

    @property
    def transformers(self) -> Transformers[SourceType, IdType]:
        """Get a dict ``{source: transformer}`` of :class:`.Transformer` instances used by this ``Translator``."""
        return self._transformers

    def _execute_fetch(
        self, task: TranslationTask[NameType, SourceType, IdType]
    ) -> TranslationMap[NameType, SourceType, IdType]:
        start = perf_counter()
        source_to_ids = task.extract_ids()

        for source in source_to_ids:
            if (transformer := self._transformers.get(source)) is not None:
                transformer.update_ids(source_to_ids[source])

        ids_to_fetch = [IdsToFetch(source, ids=ids) for source, ids in source_to_ids.items()]
        source_translations = self._fetch(ids_to_fetch, fmt=task.fmt, task_id=task.task_id)
        translation_map = self._to_translation_map(source_translations, fmt=task.fmt)

        task.add_timing("fetch", perf_counter() - start)

        return translation_map

    def _fetch(
        self,
        ids_or_sources: list[IdsToFetch[SourceType, IdType]] | set[SourceType] | None,
        fmt: Format | None = None,
        task_id: int | None = None,
    ) -> SourcePlaceholderTranslations[SourceType]:
        fmt = fmt or self._fmt
        placeholders = fmt.placeholders
        required = fmt.required_placeholders

        if self._default_fmt and ID in self._default_fmt.placeholders and ID not in placeholders:
            # Ensure that default translations can always use the ID
            placeholders = (*placeholders, ID)
            required = (*required, ID)

        if task_id is None:
            task_id = _logging.generate_task_id()

        if ids_or_sources is None or isinstance(ids_or_sources, set):
            return self.fetcher.fetch_all(
                placeholders,
                required=required,
                sources=ids_or_sources,
                task_id=task_id,
                enable_uuid_heuristics=self._enable_uuid_heuristics,
            )
        else:
            return self.fetcher.fetch(
                ids_or_sources,
                placeholders,
                required=required,
                task_id=task_id,
                enable_uuid_heuristics=self._enable_uuid_heuristics,
            )

    def _to_translation_map(
        self,
        source_translations: SourcePlaceholderTranslations[SourceType],
        fmt: Format | None = None,
    ) -> TranslationMap[NameType, SourceType, IdType]:
        return TranslationMap(
            source_translations,
            fmt=fmt or self._fmt,
            default_fmt=self._default_fmt,
            default_fmt_placeholders=self._default_fmt_placeholders,
            enable_uuid_heuristics=self._enable_uuid_heuristics,
            transformers=self._transformers,
        )

    def __repr__(self) -> str:
        more = f"fetcher={self.fetcher}" if self.online else f"cache={self.cache}"

        online = self.online
        return f"{tname(self)}({online=}: {more})"


def _handle_default(
    default_fmt: FormatType,
    default_fmt_placeholders: MakeType[SourceType, str, Any] | None,
) -> tuple[InheritedKeysDict[SourceType, str, Any], Format]:
    default_fmt = Format.parse(default_fmt)

    if not default_fmt_placeholders:
        return InheritedKeysDict(), default_fmt

    if isinstance(default_fmt_placeholders, InheritedKeysDict):
        default_placeholders = default_fmt_placeholders
    else:
        default_placeholders = InheritedKeysDict.make(default_fmt_placeholders)

    return default_placeholders, default_fmt



from collections.abc import Sequence
from typing import Any
from uuid import UUID

from id_translation.offline import TranslationMap
from id_translation.types import IdType, NameType, SourceType

from .._data_structure_io import DataStructureIO
from ..exceptions import NotInplaceTranslatableError


class SingleValueIO(DataStructureIO[IdType, NameType, SourceType, IdType]):
    """IO implementation for ``int``, ``str`` and ``UUID`` types."""

    priority = 1500

    @classmethod
    def handles_type(cls, arg: Any) -> bool:
        return isinstance(arg, (int, str, UUID))

    @classmethod
    def extract(cls, translatable: IdType, names: list[NameType]) -> dict[NameType, Sequence[IdType]]:
        if len(names) != 1:  # pragma: no cover
            raise ValueError(f"Length of {names=} must be one.")

        return {names[0]: (translatable,)}

    @classmethod
    def insert(
        cls,
        translatable: IdType,
        names: list[NameType],
        tmap: TranslationMap[NameType, SourceType, IdType],
        copy: bool,
    ) -> str:
        if not copy:  # pragma: no cover
            raise NotInplaceTranslatableError(translatable)

        return tmap[names[0]][translatable]



import json
from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import UTC, datetime, timedelta
from importlib.metadata import version as get_version
from pathlib import Path
from typing import Any, Literal, Self

from rics.strings import format_seconds as fmt_sec

from id_translation.translator_typing import CacheMissReasonType


class BaseMetadata(ABC):
    """Base implementation for Metadata types.

    Args:
        versions: Versions, e.g. ``{'python': '3.11.11'`, 'your-package': '1.0.0'}``.
        created: The time at which the metadata was originally created.
    """

    def __init__(
        self,
        versions: dict[str, str] | None = None,
        created: datetime | None = None,
    ) -> None:
        self.versions = self.get_package_versions([]) if versions is None else versions
        self.created = created or datetime.now(UTC)

    @abstractmethod
    def _to_dict(self, to_json: dict[str, Any]) -> dict[str, Any]:
        """Turn `to_json` into JSON-serializable types."""

    @classmethod
    @abstractmethod
    def _deserialize(cls, from_json: dict[str, Any]) -> dict[str, Any]:
        """Turn `from_json` into desired types."""

    @abstractmethod
    def _is_equivalent(self, other: Self) -> str:
        """Implementation-specific equivalence check."""

    def is_equivalent(self, other: Self) -> str:
        """Compute equivalency with `other`.

        Args:
            other: Another metadata instance.

        Returns:
            A string `reason_not_equivalent` or an empty string.
        """
        if not isinstance(other, self.__class__):
            return f"Expected class={self.__class__.__name__} but got {other.__class__}"

        for package, version in self.versions.items():
            other_version = other.versions.get(package)
            if other_version != version:
                return f"Expected {package}={version!r} (your environment) but got {package}={other_version!r}"

        return self._is_equivalent(other)

    def to_dict(self) -> dict[str, Any]:
        """Get a dict representation of this ``BaseMetadata``."""
        raw = self.__dict__.copy()
        kwargs = dict(
            versions=raw.pop("versions"),
            created=raw.pop("created").isoformat(),
        )
        kwargs.update(self._to_dict(raw))
        assert not raw, f"Not serialized: {raw}."  # noqa:  S101
        return kwargs

    def to_json(self) -> str:
        """Get a dict representation of this ``BaseMetadata``."""
        kwargs = self.to_dict()
        return json.dumps(kwargs, indent=4)

    @classmethod
    def from_json(cls, s: str) -> "BaseMetadata":
        """Create ``BaseMetadata`` from a JSON string `s`."""
        raw = json.loads(s)

        kwargs = dict(
            versions=raw.pop("versions"),
            created=datetime.fromisoformat(raw.pop("created")),
        )
        kwargs.update(cls._deserialize(raw))

        assert not raw, f"Not deserialized: {raw}."  # noqa:  S101
        return cls(**kwargs)

    def use_cached(
        self,
        metadata_path: Path,
        max_age: str | timedelta | None,
    ) -> tuple[Literal[True], str, None] | tuple[Literal[False], str, CacheMissReasonType]:
        """Check status of stored metadata config based a desired configuration ``self``.

        Args:
            metadata_path: Path of stored metadata.
            max_age: Maximum age of stored metadata. Pass zero to force recreation. Smaller than zero = never expire.

        Returns:
            A tuple ``(True, expires_at, None)`` or ``(False, reason, reason_type)``.
        """
        if not metadata_path.exists():
            return False, "no cache metadata found", "metadata-missing"

        stored_config = self.from_json(metadata_path.read_text())

        reason_not_equivalent = self.is_equivalent(stored_config)
        if reason_not_equivalent:
            return False, f"cached instance is not equivalent: {reason_not_equivalent}", "metadata-changed"

        if max_age is None:
            return True, "does not expire", None
        elif isinstance(max_age, str):
            delta = self._delta_from_string(max_age)
        else:
            delta = max_age

        expires_at = (stored_config.created + abs(delta)).replace(microsecond=0)
        offset = fmt_sec(round(abs(datetime.now(UTC) - expires_at).total_seconds()))

        if expires_at <= stored_config.created:
            return False, f"expired at {expires_at.isoformat()} ({offset} ago)", "too-old"
        else:
            return True, f"expires at {expires_at.isoformat()} (in {offset})", None

    @classmethod
    def _delta_from_string(cls, max_age: str) -> timedelta:
        try:
            import pandas  # noqa: PLC0415

            return pandas.Timedelta(max_age).to_pytimedelta()  # type: ignore[no-any-return]
        except ImportError:
            pass

        unit = ""
        for c in max_age:
            if c.isalpha():
                unit = c
                break

        if not unit:
            msg = f"bad {max_age=}"
            raise ValueError(msg)

        unit_multiplier: int
        match unit.lower():
            case "s":
                unit_multiplier = 1
            case "m":
                unit_multiplier = 60
            case "h":
                unit_multiplier = 60 * 60
            case "d":
                unit_multiplier = 60 * 60 * 24
            case _:
                raise ValueError(f"bad {max_age=}")

        n_units = int(max_age[: max_age.index(unit)])
        return timedelta(seconds=n_units * unit_multiplier)

    @classmethod
    def get_package_versions(cls, extra_packages: Iterable[str]) -> dict[str, str]:
        """Extract package versions using ``importlib.metadata``."""
        assert not isinstance(extra_packages, str)  # noqa: S101
        packages = ["id-translation", *extra_packages]
        return {package: get_version(package) for package in packages}



"""Types related to translation fetching."""

import typing as _t
from dataclasses import dataclass as _dataclass

from .. import types as _tt
from ..offline import types as _ot


@_dataclass(frozen=True)
class IdsToFetch(_t.Generic[_tt.SourceType, _tt.IdType]):
    """A source and the IDs to fetch from it."""

    source: _tt.SourceType
    """Where to fetch from."""
    ids: set[_tt.IdType]
    """Unique IDs to fetch translations for."""


@_dataclass(frozen=True)
class FetchInstruction(_t.Generic[_tt.SourceType, _tt.IdType]):
    """Instructions passed from an ``AbstractFetcher`` to an implementation."""

    source: _tt.SourceType
    """Where to fetch from."""
    placeholders: _ot.PlaceholdersTuple
    """All desired placeholders in preferred order."""
    required: set[str]
    """Placeholders that must be included in the response."""
    ids: set[_tt.IdType] | None
    """Unique IDs to fetch translations for."""
    task_id: int
    """Used for logging purposes."""
    enable_uuid_heuristics: bool
    """Improves matching when :py:class:`~uuid.UUID`-like IDs are in use.

    Implementations which have no UUID heuristics may silently ignore this flag.
    """

    @property
    def fetch_all(self) -> bool:
        """If ``True``, retrieve all available data."""
        return self.ids is None


FetchOperation = _t.Literal["FETCH", "FETCH_ALL"]
"""Key operation types for fetching."""
Operation = _t.Literal["INITIALIZE_SOURCES"] | FetchOperation
"""Key operation types for fetchers."""



"""Functions which perform heuristics for score functions.

See Also:
    The :class:`~.HeuristicScore` class.
"""

import collections.abc as _abc
import logging
import re as _re
import typing as _t

from rics.misc import get_by_full_name as _get_by_full_name

from .. import logging as _logging

_NOUN_TRANSFORMER_CACHE: dict[str, _abc.Callable[[str], str]] = {}
PluralToSingularArg = bool | dict[str, str] | _abc.Callable[[str], str] | str


def like_database_table(
    name: str,
    tables: _abc.Iterable[str],
    context: _t.Any,  # noqa: ARG001
    *,
    plural_to_singular: PluralToSingularArg = True,
) -> tuple[str, list[str]]:
    """Normalize `name` and `tables` to appear as base-form nouns.

    Args:
        name: A name to find a translation source.
        tables: Database tables used as possible translation sources.
        context: Ignored.
        plural_to_singular: Convert plural-form to singular form. Pass a ``dict`` to specify custom transformations,
            backed by the default transformer. See :class:`NounTransformer` for details. Set to ``False`` to disable.

            To use a custom transformer, pass a callable ``(str) -> str``, or the fully qualified name of such a
            callable. The callable will be resolved using :func:`rics.misc.get_by_full_name`, then cached.

    Returns:
        A tuple ``(normalized_name, normalized_table_names)``.

    See Also:
        * :func:`smurf_columns`

    Examples:
        Remove ID suffixes and convert a variety of plural forms to singular forms.

        >>> like_database_table("dog_id", ["dog", "dogs"], None)
        ('dog', ['dog', 'dog'])
        >>> like_database_table("city_ids", ["city", "cities"], None)
        ('city', ['city', 'city'])
        >>> like_database_table("CountryBitmask", ["Country", "COUNTRIES"], None)
        ('country', ['country', 'country'])

        Inputs are coerced to lower case.
    """
    to_singular = _get_noun_transformer(plural_to_singular)

    def _normalize_noun(noun: str) -> str:
        noun = noun.lower()
        noun = _strip_id_suffix(noun)
        noun = noun.replace("_", "")
        noun = noun.replace(".", "")
        noun = to_singular(noun)
        return noun

    return _normalize_noun(name), [_normalize_noun(t) for t in tables]


def smurf_columns(
    placeholder: str,
    columns: _abc.Iterable[str],
    table: str,
    *,
    plural_to_singular: PluralToSingularArg = False,
) -> set[str]:
    """Short-circuit `placeholder` to a matching smurf column.

    The smurf naming convention (or anti-pattern, depending on who you ask) refers the practice of including the name
    of the table in the column name, especially for the primary key ID column.

    Typical columns one might encounter are ``country.country_id`` and ``cities.city_name``. Note that, for the latter
    match to be made, you must pass ``plural_to_singular=True | dict``.

    Special handling is implemented for ``placeholder="name"``, which will match when the singular-form table name is
    found in `columns`.

    Args:
        placeholder: A :class:`~id_translation.offline.Format` placeholder.
        columns: The columns of a database table.
        table: A ``Translator`` :attr:`source <id_translation.Translator.sources>` table to which
            the `columns` (or :attr:`~id_translation.Translator.placeholders`) belong.
        plural_to_singular: Convert plural-form to singular form. Pass a ``dict`` to specify custom transformations,
            backed by the default transformer. See :class:`NounTransformer` for details. Set to ``False`` to disable.

            To use a custom transformer, pass a callable ``(str) -> str``, or the fully qualified name of such a
            callable. The callable will be resolved using :func:`rics.misc.get_by_full_name`, then cached.

    Returns:
        A single-element set ``{column}``, iff a match is found. An empty set otherwise.

    Examples:
        Default translation :class:`~id_translation.offline.Format` (``{id}:{name}``) placeholders.

        >>> smurf_columns("id", ["city_id", "city_name", "city"], "city")
        {'city_id'}

        Both plural and singular form table names are supported, but the plural-to-singular transformation must be
        explicitly enabled with ``plural_to_singular=True``.

        >>> smurf_columns(
        ...     "name", ["city_id", "city_name"], "cities", plural_to_singular=True
        ... )
        {'city_name'}

        Special handling for ``placeholder="name"`` when the table name is also a column.

        >>> smurf_columns("name", ["city_id", "city_name", "city"], "city")
        {'city'}

        As with any short-circuiting function, an empty set is returned when no match is found.

        >>> smurf_columns("id", ["dog_id", "bestie_name"], "bad_dogs")
        set()

        You may add custom mappings for irregular nouns.

        >>> smurf_columns(
        ...     "id", ["goose_id"], "geese", plural_to_singular={"geese": "goose"}
        ... )
        {'goose_id'}

    Notes:
        This function acts similarly to chained calls to :func:`value_fstring_alias`, using
        ``fstring="{context}", for_value="name"`` and ``fstring="{context}_{value}"``, but is more powerful since it is
        able to preprocess the inputs.

    See Also:
        * :func:`short_circuit`
        * :func:`like_database_table`
        * :func:`value_fstring_alias`
    """
    to_singular = _get_noun_transformer(plural_to_singular)
    table = to_singular(table.lower())

    placeholder = placeholder.lower()
    columns = {c.lower() for c in columns}

    if placeholder == "name" and table in columns:
        return {table}

    smurf = f"{table}_{placeholder}"
    for column in columns:
        if smurf == column.lower():
            return {smurf}

    return set()


def short_circuit(
    value: str,
    candidates: set[str],
    context: _t.Any,  # noqa: ARG001
    *,
    value_regex: str | _re.Pattern[str],
    target_candidate: str,
    task_id: int | None = None,
) -> set[str]:
    """Short-circuit `value` to the target candidate if the target and regex conditions are met.

    If `target_candidate` is in `candidates` and `value` matches the given `value_regex`, a single-element set
    ``{target_candidate}`` is returned which will trigger short-circuiting in the calling ``Mapper``. If either of
    these conditions fail, an empty set is returned and the mapping procedure will continue.

    Args:
        value: A value to map.
        candidates: Candidates for `value`.
        context: Always ignored, exists for compatibility.
        value_regex: A pattern match against `value`. Case-insensitive by default.
        target_candidate: The candidate to short-circuit to.
        task_id: Used for logging.

    Returns:
        A single-element set ``{target_candidate}``, iff both conditions are met. An empty set otherwise.

    Examples:
        Always match any bite victim-columns to the `humans` table (see the :ref:`translation-primer` page).

        >>> short_circuit(
        ...     "first_bite_victim",
        ...     {"humans", "animals"},
        ...     None,
        ...     value_regex=".*_bite_victim$",
        ...     target_candidate="humans",
        ... )
        {'humans'}

        Short-circuiting will only trigger if the `value_regex` matches, and the `target_candidate` is present.
    """
    candidates = set(candidates)

    if target_candidate in candidates:
        pattern = _re.compile(value_regex, flags=_re.IGNORECASE) if isinstance(value_regex, str) else value_regex
        if pattern.match(value):
            return {target_candidate}

    if _logging.ENABLE_VERBOSE_LOGGING:
        logger = logging.getLogger(__name__).getChild(short_circuit.__name__)

        pattern = _re.compile(value_regex, flags=_re.IGNORECASE) if isinstance(value_regex, str) else value_regex
        if logger.isEnabledFor(logging.DEBUG):
            if target_candidate not in candidates:
                logger.debug(
                    f"Short-circuiting failed for {value=}: The {target_candidate=} is an input candidate.",
                    extra={"task_id": task_id},
                )
            else:
                logger.debug(
                    f"Short-circuiting failed for {value=}: Does not match {pattern=}.", extra={"task_id": task_id}
                )

    return set()


def force_lower_case(
    value: str,
    candidates: _abc.Iterable[str],
    context: _t.Any,  # noqa: ARG001
) -> tuple[str, list[str]]:
    """Force lower-case in `value` and `candidates`."""
    return value.lower(), [c.lower() for c in candidates]


def value_fstring_alias(
    value: str,
    candidates: _abc.Iterable[str],
    context: _t.Any,
    *,
    fstring: str,
    for_value: str | None = None,
    **kwargs: _t.Any,
) -> tuple[str, _abc.Iterable[str]]:
    """Return a value formatted by `fstring`.

    .. note::

       This function modifies the `value`. Candidates are always returned as-is.

    Args:
        value: An element to find matches for.
        candidates: Potential matches for `value`. Not used (returned as given).
        context: Context in which the function is being called.
        fstring: The format string to use. Can use `value` and `context` as placeholders.
        for_value: If given, apply only if ``value == for_value``. When `for_value` is given, `fstring` arguments which
            do not use the `value` as a placeholder key are permitted.
        **kwargs: Additional keyword placeholders in `fstring`.

    Returns:
        A tuple ``(formatted_value, candidates)``.

    Raises:
        ValueError: If `fstring` does not contain a placeholder `'value'` and `for_value` is not given.

    Examples:
        Keys ``{value}`` and ``{context}`` are always available.

        >>> value_fstring_alias("id", ["dog_id"], "dog", fstring="{context}_{value}")
        ('dog_id', ['dog_id'])

        In cases such as these, consider using :func:`smurf_columns` instead, which will work both for ``table="dog"``
        (as above), and with ``table="dogs"``.
    """
    if not for_value and "{value}" not in fstring:
        # No longer a function of the value.
        raise ValueError(
            f"Invalid {fstring=} passed to value_fstring_alias(); does not contain {{value}}. "
            "To allow, the 'for_value' parameter must be given as well."
        )

    if for_value and value != for_value:
        return value, candidates

    return fstring.format(value=value, context=context, **kwargs), candidates


def candidate_fstring_alias(
    value: str,
    candidates: _abc.Iterable[str],
    context: _t.Any,
    *,
    fstring: str,
    **kwargs: _t.Any,
) -> tuple[str, _abc.Iterable[str]]:
    """Return candidates formatted by `fstring`.

    .. note::

       This function modifies the `candidates`. The `value` is always returned as-is.

    Args:
        value: An element to find matches for. Not used (returned as given).
        candidates: Potential matches for `value`.
        context: Context in which the function is being called.
        fstring: The format string to use. Can use `value`, `context`, and elements of `candidates` as placeholders.
        **kwargs: Additional keyword placeholders in `fstring`.

    Returns:
        A tuple ``(value, formatted_candidates)``.

    Raises:
        ValueError: If `fstring` does not contain a placeholder `'candidate'`.
    """
    if "{candidate}" not in fstring:
        raise ValueError(f"Invalid {fstring=} passed to candidate_fstring_alias(); does not contain {{candidate}}.")

    return value, [fstring.format(value=value, candidate=c, context=context, **kwargs) for c in candidates]


def _strip_id_suffix(string: str) -> str:
    for suffix in "id", "ids", "bitmask":
        sz = len(suffix)
        if len(string) > sz and string.endswith(suffix):
            string = string[:-sz]

    string = string.removesuffix("_")

    return string


class NounTransformer:
    """Naive utility class for transforming nouns to singular form.

    This class performs simple heuristics to convert nouns commonly used as database table names. It will quickly break
    either if given nouns that are already on singular form, or are not trivially convertible (see
    :attr:`PLURAL_TO_SINGULAR_SUFFIXES`) to singular form.

    .. note::
       For more complex use cases, consider using a language-processing framework such as
       `inflect (PyPI) <https://pypi.org/project/inflect/>`_ instead.

    Pass ``plural_to_singular=<fully-qualified-name>`` to use your implementation in any function that accepts a
    `plural_to_singular`-argument.

    .. code-block:: python

       def my_transform(plural: str) -> str:
           import inflect

           p = inflect.engine()
           return p.singular_noun(plural)


       smurf_columns(..., plural_to_singular="__main__.my_transform")

    Examples:
        >>> nt = NounTransformer(custom={"geese": "goose"})
        >>> nt("city"), nt("cities")
        ('city', 'city')
        >>> nt("country"), nt("countries")
        ('country', 'country')
        >>> nt("geese"), nt("goose"), nt("species")
        ('goose', 'goose', 'species')

        May break when given a noun that is already singular.

        >>> nt("bus"), nt("news")
        ('bu', 'new')

        See :attr:`PLURAL_TO_SINGULAR_SUFFIXES` for affected suffixes.

    Notes:
        This is not :class:`~id_translation.transform.types.Transformer` implementation, in spite of the name.
    """

    IRREGULARS: _t.ClassVar[dict[str, str]] = {
        "species": "species",
        # Not really irregular
        "phases": "phase",
        "exercises": "exercise",
    }
    """Known irregular plural-to-singular transformations."""

    # https://wordtoolbox.com/nouns-ending-with/<letter-combination>
    PLURAL_TO_SINGULAR_SUFFIXES: tuple[tuple[str, str], ...] = (
        ("ies", "y"),  # cit[ies] -> cit[y]
        ("ives", "ife"),  # l[ives] -> l[ife]
        ("ves", "f"),  # hal[ves] -> hal[f]
        ("oes", "o"),  # tomat[oes] -> tomat[o]
        ("hes", "h"),  # has[hes] -> has[h]
        ("ses", "s"),  # iri[ses] -> iri[s]
        ("xes", "x"),  # bo[xes] -> bo[x]
        ("s", ""),  # catch-all
    )
    """Plural-to-singular suffix mappings."""

    def __init__(self, custom: dict[str, str] | None = None) -> None:
        if custom is None:
            custom = {}
        self._pre = {**self.IRREGULARS, **custom}

    def __call__(self, noun: str) -> str:
        """Convert to singular form."""
        singular = self._pre.get(noun)
        return self._to_singular(noun) if singular is None else singular

    @classmethod
    def _to_singular(cls, plural: str) -> str:
        for suffix, replacement in cls.PLURAL_TO_SINGULAR_SUFFIXES:
            if plural.endswith(suffix):
                return plural[: -len(suffix)] + replacement
        return plural


def _get_noun_transformer(plural_to_singular: PluralToSingularArg) -> _abc.Callable[[str], str]:
    """Returns a NOOP if `plural_to_singular` is ``False``."""
    if plural_to_singular is False:
        return lambda noun: noun

    if isinstance(plural_to_singular, str):
        if (transformer := _NOUN_TRANSFORMER_CACHE.get(plural_to_singular)) is not None:
            return transformer

        transformer = _get_by_full_name(plural_to_singular)
        _NOUN_TRANSFORMER_CACHE[plural_to_singular] = transformer
        return transformer  # type: ignore[no-any-return]

    if callable(plural_to_singular):
        return plural_to_singular

    if plural_to_singular is True:
        plural_to_singular = {}
    return NounTransformer(plural_to_singular)



import logging
import warnings
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from time import perf_counter
from typing import TYPE_CHECKING, Any, get_args
from uuid import UUID

from rics.misc import tname
from rics.strings import format_seconds as fmt_sec

from .. import _uuid_utils
from ..exceptions import TooManyFailedTranslationsError
from ..mapping.types import UserOverrideFunction
from ..offline import Format, TranslationMap
from ..types import IdType, IdTypes, Names, NameToSource, NameType, NameTypes, SourceType, Translatable
from ._map import MappingTask

LOGGER = logging.getLogger("id_translation.Translator")

NUM_SAMPLE_IDS = 10

if TYPE_CHECKING:
    from .._translator import Translator


class TranslationTask(MappingTask[NameType, SourceType, IdType]):
    """Ephemeral class for performing a single translation task on a `translatable`."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
        fmt: Format,
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        *,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        copy: bool = True,
        max_fails: float = 1.0,
        reverse: bool = False,
        enable_uuid_heuristics: bool = False,
        event_key: str,
        task_id: int | None = None,
    ) -> None:
        super().__init__(
            caller,
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            task_id=task_id,
        )

        if not (0.0 <= max_fails <= 1):
            raise ValueError(f"Argument {max_fails=} is not a valid fraction")

        self.fmt = fmt
        self.copy = copy
        self.max_fails = max_fails
        self.reverse = reverse
        self.parent: Translatable[NameType, IdType] | None = None

        self.enable_uuid_heuristics = enable_uuid_heuristics

        self._names_without_ids: set[NameType] = set()
        self._event_key = event_key

        self._num_ids: dict[SourceType, int] | None = None
        self._seconds = float("nan")

    @property
    def io_names(self) -> list[NameType]:
        """Names for which IDs should be extracted from the `translatable`."""
        # Preserve input order for names, if given. These names may be repeated.
        if self.names_from_user is None:
            names = self.names_to_translate
        else:
            names = self.names_from_user

        return [n for n in names if n not in self._names_without_ids]

    def extract_ids(self) -> dict[SourceType, set[IdType]]:
        """Extract IDs to fetch from the translatable."""
        start = perf_counter()
        name_to_source = self.name_to_source
        source_to_ids: dict[SourceType, set[IdType]] = defaultdict(set)

        float_names: list[NameType] = []
        num_coerced = 0
        ids: Sequence[IdType]
        for name, ids in self._extract_ids().items():
            if len(ids) == 0:
                continue

            all_ids = source_to_ids[name_to_source[name]]

            was_coerced: bool = False
            if isinstance(ids[0], float):
                # Float IDs aren't officially supported, but is common when using Pandas since int types cannot be NaN.
                # This is sometimes a problem for the built-in set (see https://github.com/numpy/numpy/issues/9358), and
                # for several database drivers.

                try:
                    ids, n_new = self._coerce_float_to_int(ids)  # noqa: PLW2901
                    num_coerced += n_new  # Somewhat inaccurate; includes repeat IDs from other names
                    float_names.append(name)
                    was_coerced = True
                except (TypeError, ImportError):
                    pass

            if not was_coerced and self.enable_uuid_heuristics:
                ids = _uuid_utils.try_cast_many(ids)  # noqa: PLW2901

            all_ids.update(ids)

        if num_coerced > 100:  # pragma: no cover  # noqa: PLR2004
            types = f"({', '.join(t.__name__ for t in get_args(IdTypes))})"
            warnings.warn(
                f"To ensure proper fetcher operation, {num_coerced} float-type IDs have been coerced to int. "
                f"Enforcing supported data types {types} for IDs in your {self.type_name}-data may improve performance."
                f" Affected names ({len(float_names)}): {float_names}.",
                stacklevel=3,
            )

        self._num_ids = {source: len(ids) for source, ids in source_to_ids.items()}
        self.add_timing("extract", perf_counter() - start)

        return source_to_ids

    @classmethod
    def _coerce_float_to_int(cls, ids: Sequence[float]) -> tuple[Iterable[int], int]:
        from numpy import isnan, unique  # noqa: PLC0415

        arr = unique(ids)
        keep_mask = ~isnan(arr, casting="no")
        arr = arr[keep_mask]
        arr = arr.astype(int, copy=False)
        return arr, keep_mask.sum()

    def log_key_event_enter(self) -> None:
        """Emits the enter message."""
        if not LOGGER.isEnabledFor(logging.DEBUG):
            return

        type_name = self.type_name
        names = self.names_from_user
        ignore_names = self.ignore_names

        name_info = "Derive based on type" if names is None else repr(names)
        if ignore_names is not None:
            name_info += f", excluding those given by {ignore_names=}"

        LOGGER.debug(
            msg=f"Begin translation of {type_name}-type data. Names to translate: {name_info}.",
            extra=dict(
                task_id=self.task_id,
                event_key=self._event_key + ":enter",
                # Task-specific
                sources=self.caller.sources,
                online=self.caller.online,
                translatable_type=self.full_type_name,
                names=names,
                ignore_names=tname(ignore_names, prefix_classname=True) if callable(ignore_names) else ignore_names,
                copy=self.copy,
                reverse=self.reverse,
            ),
        )

    def finished(self) -> None:
        self._seconds = perf_counter() - self._start
        overhead = self._seconds - sum(self._timings.values())
        self.add_timing("overhead", overhead)

        self._log_performance()

    def _log_performance(self) -> None:
        if not LOGGER.isEnabledFor(logging.DEBUG):
            return

        durations_ms = self.get_timings_ms()
        LOGGER.debug(
            f"Performance counters [ms]: {durations_ms}",
            extra={"task_id": self.task_id, "durations_ms": durations_ms},
        )

    def log_key_event_exit(self) -> None:
        """Emits the exit message."""
        if not LOGGER.isEnabledFor(logging.INFO):
            return

        num_ids = sum(self.num_ids.values())

        in_place = "" if self.copy else "in-place "
        ids = f"{num_ids} unique ID" + ("" if num_ids == 1 else "s")
        names = "1 name" if len(self.names_to_translate) == 1 else f"{len(self.names_to_translate)} names"
        msg = f"Finished {in_place}translation of {ids} ({names}) in {self.type_name} in {fmt_sec(self._seconds)}."

        LOGGER.info(
            msg=msg,
            extra=dict(
                task_id=self.task_id,
                event_key=self._event_key + ":exit",
                seconds=self._seconds,
                # Task-specific
                durations_ms=self.get_timings_ms(),
                num_ids=self.num_ids,
                online=self.caller.online,
                translatable_type=self.full_type_name,
                io_type="{cls.__module__}.{cls.__name__}".format(cls=type(self.io)),
                copy=self.copy,
                reverse=self.reverse,
                name_to_source=self.name_to_source,
            ),
        )

    @property
    def num_ids(self) -> dict[SourceType, int]:
        """Unique ID count per source."""
        if self._num_ids is None:
            self._num_ids = self._compute_num_ids()

        return self._num_ids

    def _compute_num_ids(self) -> dict[SourceType, int]:
        """Compute number of IDs per source.

        This is done by :meth:`extract_ids` when the Translator is online. ID extraction isn't needed when IDs are
        stored offline, but we still want the counts when INFO-logging is enabled (see :meth:`log_key_event_exit`).
        """
        source_to_ids: dict[SourceType, set[IdType]] = defaultdict(set)
        for name, ids in self._extract_ids().items():
            if isinstance(ids[0], float):
                try:  # noqa: SIM105
                    ids, _ = self._coerce_float_to_int(ids)  # noqa: PLW2901
                except (TypeError, ImportError):
                    pass
            source = self.name_to_source[name]
            source_to_ids[source].update(ids)

        return {source: len(ids) for source, ids in source_to_ids.items()}

    def insert(
        self, translation_map: TranslationMap[NameType, SourceType, IdType]
    ) -> Translatable[NameType, str] | None:
        """Insert translated IDs into the `translatable`, based on data retrieved by the fetcher."""
        start = perf_counter()
        copy = self.copy

        translation_map.reverse_mode = self.reverse
        try:
            result = self.io.insert(
                self.translatable,
                names=self.io_names,
                tmap=translation_map,
                copy=copy,
            )
        finally:
            translation_map.reverse_mode = False

        self.add_timing("insert", perf_counter() - start)
        return result if copy else None

    def verify(self, tmap: TranslationMap[NameType, SourceType, IdType]) -> None:
        """Verify translations.

        Performs translation with pre-defined formats, counting the number of IDs which are (and aren't) known to the
        translation map.
        """
        max_fails = self.max_fails
        if not (max_fails < 1.0 or LOGGER.isEnabledFor(logging.DEBUG)):
            return

        start = perf_counter()
        name_to_ids = self._name_to_ids_in_order()
        translations = tmap.to_translations()

        for name, ids in name_to_ids.items():
            source = tmap.name_to_source[name]
            magic_dict = translations[source]

            if self.reverse == tmap.reverse_mode:
                is_known = magic_dict.real_contains
            else:
                is_known = {*magic_dict.real.values()}.__contains__

            is_missing = [not is_known(idx) for idx in ids]
            n_untranslated = sum(is_missing)
            if n_untranslated == 0:
                continue
            n_total = len(ids)
            f_untranslated = n_untranslated / n_total

            sample_ids = self._get_untranslated_ids(ids, is_missing_mask=is_missing)

            fail = f_untranslated > max_fails

            if not (fail or LOGGER.isEnabledFor(logging.DEBUG)):
                continue

            message = (
                f"Failed to translate {n_untranslated}/{n_total} ({f_untranslated:.2%}{{reason}}) of IDs "
                f"for {name=} using source={source!r}. Sample IDs: {sample_ids}."
            )
            extra = {
                "task_id": self.task_id,
                "name_of_ids": name,
                "source": source,
                "n_untranslated": n_untranslated,
                "n_total": n_total,
                "max_fails": max_fails,
                "sample_ids": _json_safe_types(sample_ids),
            }

            if fail:
                message = message.format(reason=f" > max_fails={max_fails:.2%}")
                LOGGER.error(message, extra=extra)
                raise TooManyFailedTranslationsError(message)

            message = message.format(reason=f" <= max_fails={max_fails:.2%}")
            LOGGER.debug(message, extra=extra)

        self.add_timing("verify", perf_counter() - start)

    def _name_to_ids_in_order(self) -> dict[NameType, Sequence[Any]]:
        name_to_ids: dict[NameType, Sequence[Any]] = self._extract_ids()
        if self.enable_uuid_heuristics:
            name_to_ids = {name: _uuid_utils.try_cast_many(name_to_ids[name]) for name in name_to_ids}
        return name_to_ids

    def _extract_ids(self) -> dict[NameType, Sequence[IdType]]:
        name_to_ids: dict[NameType, Sequence[IdType]] = self.io.extract(self.translatable, self.io_names)

        name_to_ids = {name: ids for name, ids in name_to_ids.items() if len(ids) > 0}

        if not self._names_without_ids:
            self._names_without_ids = set(self.io_names).difference(name_to_ids)
        return name_to_ids

    @staticmethod
    def _get_untranslated_ids(ids: Sequence[IdType], *, is_missing_mask: Sequence[bool]) -> list[IdType]:
        seen = set()
        retval = []

        for idx, is_missing in zip(ids, is_missing_mask, strict=True):
            if not is_missing or idx in seen:
                continue

            seen.add(idx)
            retval.append(idx)
            if len(retval) == NUM_SAMPLE_IDS:
                break

        return retval


def _json_safe_types(items: list[Any]) -> list[Any]:
    converters: dict[type[Any], Callable[[Any], Any]] = {UUID: str}

    try:
        from numpy import floating, integer  # noqa: PLC0415

        converters[integer] = int
        converters[floating] = float
    except ImportError:
        pass

    def convert(item: Any) -> Any:
        for cls, converter in converters.items():
            if isinstance(item, cls):
                return converter(item)
        return item

    return [convert(e) for e in items]



import os
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, TypeVar

from rics.collections.misc import as_list

from id_translation.offline import TranslationMap
from id_translation.types import IdType, NameType, SourceType

from .._data_structure_io import DataStructureIO
from ..exceptions import NotInplaceTranslatableError

try:
    from numpy import array, ndarray
except ImportError:

    class ndarray:  # type: ignore  # noqa
        """Dummy implementation; numpy is not available."""

    array = ndarray  # type: ignore[assignment]

SequenceT = TypeVar("SequenceT", list, ndarray, tuple)  # type: ignore[type-arg]  # TODO: Higher-Kinded TypeVars


class SequenceIO(DataStructureIO[SequenceT, NameType, SourceType, IdType]):
    """IO implementation for ``list``, ``tuple`` and  ``numpy.array`` types."""

    priority = 1100

    if not TYPE_CHECKING and os.environ.get("SPHINX_BUILD") == "true":
        SequenceT = SequenceT  # type: ignore[attr-defined]
        """Generic sequence type.

        This property does not exist at runtime. The element type cannot be parameterized due to limitations in the
        ``TypeVar`` implementation.
        """

    @classmethod
    def handles_type(cls, arg: Any) -> bool:
        return isinstance(arg, (list, ndarray, tuple))

    @classmethod
    def extract(cls, translatable: SequenceT, names: list[NameType]) -> dict[NameType, Sequence[IdType]]:
        verify_names(len(translatable), names)
        return (
            {names[0]: as_list(translatable)}
            if len(names) == 1
            else {n: as_list(r) for n, r in zip(names, translatable, strict=True)}
        )

    @classmethod
    def insert(
        cls,
        translatable: SequenceT,
        names: list[NameType],
        tmap: TranslationMap[NameType, SourceType, IdType],
        copy: bool,
    ) -> SequenceT | None:
        verify_names(len(translatable), names)
        t = translate_sequence(translatable, names, tmap)

        ctor: Callable[[list[str | None]], SequenceT]
        if copy:
            if isinstance(translatable, ndarray):
                ctor = array
            else:
                ctor = type(translatable)
            return ctor(t)

        try:
            translatable[:] = t[:]  # type: ignore
            return None
        except TypeError as e:
            raise NotInplaceTranslatableError(translatable) from e


def translate_sequence(
    s: SequenceT, names: list[NameType], tmap: TranslationMap[NameType, SourceType, IdType]
) -> list[str]:
    """Return a translated copy of the sequence `s`."""
    if len(names) == 1:
        magic_dict = tmap[names[0]]
        return [magic_dict[i] for i in s]

    return [  # TODO resolve_io med cache
        translate_sequence(element, [name], tmap)  # type: ignore
        if SequenceIO.handles_type(element)
        else tmap[name][element]
        for name, element in zip(names, s, strict=True)
    ]


def verify_names(data_len: int, names: list[NameType]) -> None:  # pragma: no cover
    """Verify that the length of names is either 1 or equal to the length of the data."""
    num_names = len(names)
    if num_names not in {1, data_len}:
        raise ValueError(
            f"Number of names {len(names)} must be 1 or equal to the length of the data ({data_len}) to "
            f"translate, but got {names=}."
        )
